import hashlib, json, sys
# import datetime

def hashMe(msg=""):

    if type(msg)!=str:
        msg = json.dumps(msg,sort_keys=True)
        
    if sys.version_info.major == 2:
        return unicode(hashlib.sha256(msg).hexdigest(),'utf-8')

    else:
        return hashlib.sha256(str(msg).encode('utf-8')).hexdigest()

def updatecontract(txn, contract):
       # If the transaction is valid, then update the contract
    contract = contract.copy() # As dictionaries are mutable, let's avoid any confusion by creating a working copy of the data.
    for key in txn:
        if key in contract.keys():
            contract[key] += txn[key]
        else:
            contract[key] = txn[key]
    return contract

minimum_order_quantity = 10
Shares_offering = 1000

def isValidTxn(txn,contract):
  
    # if sum(txn.values()) is not 0:
    #     return False
    if (txn.get('Number_of_Shares')<minimum_order_quantity):
        return False  
    
    if (txn.get('Number_of_Shares')>Shares_offering):
        return False  

    return True

contract = {'Investor':"AA", 'Issuer':"ZZ",'Amount':1000,'Number_of_Shares':100}  # Define the initial contract
genesisBlockTxns = [contract]
genesisBlockContents = {'blockNumber':0,'parentHash':None,'txnCount':1,'txns':genesisBlockTxns}
genesisHash = hashMe( genesisBlockContents )
genesisBlock = {'hash':genesisHash,'contents':genesisBlockContents}
genesisBlockStr = json.dumps(genesisBlock, sort_keys=True)

chain = [genesisBlock]

def makeBlock(txns,chain):
    parentBlock = chain[-1]
    parentHash  = parentBlock['hash']
    blockNumber = parentBlock['contents']['blockNumber'] + 1
    # Timestamp = datetime.datetime.now()
    txnCount    = len(txns)
    blockContents = {'blockNumber':blockNumber,'parentHash':parentHash,
                     'txnCount':len(txns),'txns':txns}
    blockHash = hashMe( blockContents )
    block = {'hash':blockHash,'contents':blockContents}
    
    return block

txnbuffer = [{'Investor':"AA", 'Issuer':"ZZ",'Amount':(i+1000),'Number_of_Shares':(i+100)} for i in range(5)]
a = {'Investor':"AA", 'Issuer':"ZZ",'Amount':1000,'Number_of_Shares':10020}
txnbuffer.append(a)

blockSizeLimit = 2

while len(txnbuffer) > 0:
    bufferStartSize = len(txnbuffer)
    
    txnList = []
    while (len(txnbuffer) > 0) & (len(txnList) < blockSizeLimit):
        newTxn = txnbuffer.pop()
        validTxn = isValidTxn(newTxn,contract)
        
        if validTxn:
            txnList.append(newTxn)
            contract = updatecontract(newTxn,contract)
        else:
            sys.stdout.flush()
            continue
        
    ## Make a block
    myBlock = makeBlock(txnList,chain)
    chain.append(myBlock) 

    print(chain)

def checkBlockHash(block):
    
    expectedHash = hashMe( block['contents'] )
    if block['hash']!=expectedHash:
        raise Exception('Hash does not match contents of block %s'%
                        block['contents']['blockNumber'])
    return


def checkBlockValidity(block,parent,contract):    
    # We want to check the following conditions:
    # - Each of the transactions are valid updates to the system contract
    # - Block hash is valid for the block contents
    # - Block number increments the parent block number by 1
    # - Accurately references the parent block's hash
    parentNumber = parent['contents']['blockNumber']
    parentHash   = parent['hash']
    blockNumber  = block['contents']['blockNumber']
    # Check transaction validity; throw an error if an invalid transaction was found.
    for txn in block['contents']['txns']:
        if isValidTxn(txn,contract):
            contract = updatecontract(txn,contract)
        else:
            raise Exception('Invalid transaction in block %s: %s'%(blockNumber,txn))

    checkBlockHash(block) # Check hash integrity; raises error if inaccurate

    if blockNumber!=(parentNumber+1):
        raise Exception('Hash does not match contents of block %s'%blockNumber)

    if block['contents']['parentHash'] != parentHash:
        raise Exception('Parent hash not accurate at block %s'%blockNumber)
    
    return contract

def checkChain(chain):
    # Work through the chain from the genesis block (which gets special treatment), 
    #  checking that all transactions are internally valid,
    #    that the transactions do not cause an overdraft,
    #    and that the blocks are linked by their hashes.
    # This returns the contract as a dictionary of accounts and balances,
    #   or returns False if an error was detected

    
    ## Data input processing: Make sure that our chain is a list of dicts
    if type(chain)==str:
        try:
            chain = json.loads(chain)
            assert( type(chain)==list)
        except:  # This is a catch-all, admittedly crude
            return False
    elif type(chain)!=list:
        return False
    
    contract = {}
    
    # We want to check the following conditions:
    # - Each of the transactions are valid updates to the system contract
    # - Block hash is valid for the block contents

    for txn in chain[0]['contents']['txns']:
        contract = updatecontract(txn,contract)
    checkBlockHash(chain[0])
    parent = chain[0]
    
    ## Checking subsequent blocks: These additionally need to check
    #    - the reference to the parent block's hash
    #    - the validity of the block number
    for block in chain[1:]:
        contract = checkBlockValidity(block,parent,contract)
        parent = block
        
    return contract


import copy
nodeBchain = copy.copy(chain)
nodeBtxns  = [txnbuffer for i in range(5)]
nodeBtxns = [{'Investor':"AA", 'Issuer':"ZZ",'Amount':(i+1000),'Number_of_Shares':(100*i+100)} for i in range(5)]
newBlock   = makeBlock(nodeBtxns,nodeBchain)


print("Blockchain on Node A is currently %s blocks long"%len(chain))

try:
    print("New Block Received; checking validity...")
    contract = checkBlockValidity(newBlock,chain[-1],contract) # Update the contract- this will throw an error if the block is invalid!
    chain.append(newBlock)
except:
    print("Invalid block; ignoring and waiting for the next block...")

print("Blockchain on Node A is now %s blocks long"%len(chain))           

